#!/usr/bin/perl
##############################################################################
#    This program is free software: you can redistribute it and/or modify    #
#    it under the terms of the GNU General Public License as published by    #
#    the Free Software Foundation, either version 3 of the License, or       #
#    (at your option) any later version.				     #
#									     #
#    This program is distributed in the hope that it will be useful,   	     #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of	     #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     #
#    GNU General Public License for more details.			     #
#									     #
#    You should have received a copy of the GNU General Public License	     #
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.   #
##############################################################################

# Identify Genomic SV breakpoints underlying fusion junctions
use strict;
use warnings;
no warnings 'recursion';
use Getopt::Std;
use FindBin qw($Bin);
use lib "$FindBin::Bin/lib";

my $version='BreakTrans-0.0.6';
my %opts = (
	    g=>"database/Human.Mar2006.RefSeqGenes.tab",
	    l=>200,
	    z=>100000,
	    h=>1,
	    C=>'database/Human.Mar2006.chainSelf.tab'
	   );
my %opts1;

getopts('g:l:o:C:h:z:',\%opts);
die("
Identify Genomic SV breakpoints underlying fusion junctions
Usage:   BreakTrans.pl <genomic breakpoints> <gene fusion breakpoints>\n
Options:
         -o STR   Only analysis the specified chromosome
         -g FILE  Use UCSC Human gene table at $opts{g}
         -l INT   Proximity of breakpoints to gene annotations [$opts{l}] bp
         -C       Use UCSC selfchain alignment FILE, default: $opts{C}
         -h FLOAT Normalized score cutoff in selfchain alignment, default: $opts{h}
         -z INT   Self-chain region cutoff, default: $opts{z} bp
Version: $version
\n") unless (@ARGV);

my (%BK1s,%BK2s,%BKs,%chrs);
my %edges;  #breakpoint graph
my %edges_cn;  #copy number
my $header;
my ($f_SV,$f_fusion)=@ARGV;

my $options='';
foreach my $opt(keys %opts1){
  $options.='|'.$opt.$opts1{$opt};
  $opts{$opt}=$opts1{$opt};
}

#Read in a SV file and create genomic breakpoint graph
open(IN,"<$f_SV") || die "unable to open $f_SV\n";
while(<IN>){
  chomp;
  my @u=split /\s+/;
  my ($chr,$start,$ori1,$chr2,$end,$ori2,$type,$size,$score,$nreads,$nread_lib,@cn)=@u;
  next if(defined $opts{o} && $chr ne $opts{o});
  $chr=~s/chr//;
  $chr2=~s/chr//;

  next unless($start=~/^\d+/ && $end=~/^\d+/ && defined $chr);

  $BK1s{$chr}{$start}++;
  $BK2s{$chr2}{$end}++;
  $BKs{$chr}{$start}++;
  $BKs{$chr2}{$end}++;
  $chrs{$chr}++; $chrs{$chr2}++;

  #Add breakpoint edges (undirected)
  my $strandori;  # +-,++,-+,--, first: strand into the breakpoint, second: strand out the breakpoint

  my ($nposreads1,$nposreads2,$nnegreads1,$nnegreads2);
  if($ori1=~/\d+[\+\-]\d+[\+\-]/ && $ori2=~/\d+[\+\-]\d+[\+\-]/){  #BreakDancer format
    ($nposreads1)=($ori1=~/(\d+)\+/); $nposreads1=0 if(!defined $nposreads1);
    ($nnegreads1)=($ori1=~/(\d+)\-/); $nnegreads1=0 if(!defined $nnegreads1);
    ($nposreads2)=($ori2=~/(\d+)\+/); $nposreads2=0 if(!defined $nposreads2);
    ($nnegreads2)=($ori2=~/(\d+)\-/); $nnegreads2=0 if(!defined $nnegreads2);

    my ($bkori1,$nreads1);
    my ($bkori2,$nreads2);

    if($nposreads1>$nnegreads1){
      $bkori1='+';
      $nreads1=$nposreads1;
    }
    else{
      $bkori1='-';
      $nreads1=$nnegreads1;
    }

    if($nposreads2>$nnegreads2){
      $bkori2='+';
      $nreads2=$nposreads2;
    }
    else{
      $bkori2='-';
      $nreads2=$nnegreads2;
    }

    if($type eq 'CTX' || $chr ne $chr2){
      if($bkori1 eq '+' && $bkori2 eq '-'){
	$strandori='++';
      }
      elsif($bkori1 eq '+' && $bkori2 eq '+'){
	$strandori='+-';
      }
      elsif($bkori1 eq '-' && $bkori2 eq '+'){
	$strandori='--';
      }
      elsif($bkori1 eq '-' && $bkori2 eq '-'){
	$strandori='-+';
      }
      else{}
    }
    elsif($type eq 'INV'){  #  ->|   <-|,  |<-   |->
      my $bkori;
      if($bkori1 eq $bkori2){
	$strandori=$bkori1 . &RevOri($bkori1);
      }
      else{
	$strandori=($nreads1>$nreads2)?($bkori1 . &RevOri($bkori1)):($bkori2 . &RevOri($bkori2));
      }
    }
    elsif($type eq 'ITX'){  #   |<-   <-|
      $strandori='--';
    }
    else{  # DEL, INS,  ->|    |->
      $strandori='++';
    }
  }
  elsif($ori1=~/\d+[\+\-]/ && $ori2=~/\d+[\+\-]/){  #Squaredancer format
    if($ori1=~/\+/ && $ori2=~/\-/){
      $strandori='++';
    }
    elsif($ori1=~/\+/ && $ori2=~/\+/){
      $strandori='+-';
    }
    elsif($ori1=~/\-/ && $ori2=~/\+/){
      $strandori='--';
    }
    elsif($ori1=~/\-/ && $ori2=~/\-/){
      $strandori='-+';
    }
    else{}
    ($nnegreads1)=($ori1=~/(\d+)\-/); $nnegreads1=0 if(!defined $nnegreads1);
    ($nposreads2)=($ori2=~/(\d+)\+/); $nposreads2=0 if(!defined $nposreads2);
    ($nnegreads2)=($ori2=~/(\d+)\-/); $nnegreads2=0 if(!defined $nnegreads2);   
  }
  elsif($ori1=~/^[\+\-]$/ && $ori2=~/^[\+\-]$/){
    $strandori=$ori1 . $ori2;
  }
  else{  #other format
  }

  next unless(defined $strandori);

  #Create breakpoint edges
  my ($bp1,$bp2)=(join(":",$chr,$start),join(":",$chr2,$end));
  $edges{$bp1}{$bp2}{$strandori}=$nreads;  #$score;
  #Reverse-complemented edges
  $strandori=~tr/\+\-/\-\+/;
  my $rev_strandori=reverse($strandori);
  $edges{$bp2}{$bp1}{$rev_strandori}=$nreads;  #$score;
}
close(IN);

#Read in the fusion file

my %RBKs;
my @Splicing;
open(IN,"<$f_fusion") || die "unable to open $f_fusion\n";
while(<IN>){
  next if(/^\#/);
  chomp;
  my ($chr,$start,$chr2,$end,@extra)=split /\s+/;
  next if(defined $opts{o} && $chr ne $opts{o});

  $chr=~s/chr//;
  $chr2=~s/chr//;
  next unless($start=~/^\d+/ && $end=~/^\d+/ && defined $chr);

  my $sp=join("\t",$chr,$start,$chr2,$end,@extra);
  push @Splicing,$sp;
  $RBKs{$chr}{$start}++;
  $RBKs{$chr2}{$end}++;
  $chrs{$chr}++;
  $chrs{$chr2}++;
}

my %ABKs;
my %SCTBKs;    #selfchain annotations overlap transcripts that contains the breakpoints

foreach my $chr(keys %chrs){
  next if($chr=~/^NT/ || $chr=~/^MT/);
  next if(defined $opts{o} && $chr ne $opts{o});
  my $annot=&ReadUCSCGeneAnnotation($chr);

  #Select all transcripts that overlap with the breakpoint
  my @gposes=keys %{$BKs{$chr}} if(defined $BKs{$chr});
  my @rposes=keys %{$RBKs{$chr}} if(defined $RBKs{$chr});
  my @poses=sort {$a<=>$b} (@gposes,@rposes);
  my @txEnds=sort {$a<=>$b} keys %{$$annot{$chr}};
  foreach my $pos(@poses){
    while (@txEnds>0 && $pos>$txEnds[0]+$opts{l}){
      shift @txEnds;
    }
    next unless(@txEnds>0);
    foreach my $txEnd(@txEnds){
      foreach my $start(keys %{$$annot{$chr}{$txEnd}}){
	if($pos>$start-$opts{l}){
	  foreach my $transcript(@{$$annot{$chr}{$txEnd}{$start}}){
	    push @{$ABKs{$chr}{$pos}},$transcript;
	  }
	}
      }
    }
  }
}

my %GeneBKs;  #Associate genomic breakpoints with each gene
foreach my $chr(keys %chrs){
  next unless (defined $BKs{$chr});
  foreach my $pos(sort {$a<=>$b} keys %{$BKs{$chr}}){
    my $annot=&GetBKGeneAnnotation($chr,$pos);
    push @{$GeneBKs{$annot->{gene}}},join(':',$chr,$pos) if(defined $annot);
  }
}

#Add reference adjacency for breakends within the same gene
foreach my $gene(keys %GeneBKs){
  my @loci=@{$GeneBKs{$gene}};
  my $ploc=shift @loci;
  foreach my $loc(@loci){
    $edges{$ploc}{$loc}{'++'}=0;  #0 read count reserved for reference adjacency
    $edges{$loc}{$ploc}{'--'}=0;  #0 read count reserved for reference adjacency
    $ploc=$loc;
  }
}

#Read in fusion breakpoints
my %Rchrs;
my %RCBKs;
my %fusion;
my %gBKPath;
my @confirmed;
foreach my $sp(@Splicing){
  my ($chr,$start,$chr2,$end,$source,@extra)=split /\s+/,$sp;
  $chr=~s/chr//;
  $chr2=~s/chr//;

  my $RNAbp1=&GetBKGeneAnnotation($chr,$start);
  my $RNAbp2=&GetBKGeneAnnotation($chr2,$end);
  next unless (defined $RNAbp1 && defined $RNAbp2);

  my $DNAbp1=&GetInnerDNAbp($RNAbp1,$start);
  my $DNAbp2=&GetInnerDNAbp($RNAbp2,$end);
  next unless (defined $DNAbp1 && $DNAbp2);

  if($RNAbp1->{gene} ne $RNAbp2->{gene}   #fusion
    ){

    my @pathstr=();

    my %visited;
    my @paths=&getPath(\%edges,\%visited,$RNAbp1->{strand},$DNAbp1,$DNAbp2,$RNAbp2->{strand});
    if(@paths){
      $fusion{$sp}=$RNAbp1->{gene}.'>'.$RNAbp2->{gene};
      foreach my $p(@paths){
	push @pathstr, &SimplePath($DNAbp1 . $RNAbp1->{strand}, $p);
      }
    }
    my @paths_rev=&getPath(\%edges,\%visited,$RNAbp2->{strand},$DNAbp2,$DNAbp1,$RNAbp1->{strand});
    if(@paths_rev){
      $fusion{$sp}=$RNAbp2->{gene}.'>'.$RNAbp1->{gene};
      foreach my $p(@paths_rev){
	push @pathstr, &SimplePath($DNAbp2 . $RNAbp2->{strand}, $p);
      }
    }

    if(@pathstr){
      push @confirmed,$sp;
      $gBKPath{$sp}=join(',',@pathstr);
      $Rchrs{$chr}++;
      $Rchrs{$chr2}++;
      $RCBKs{$chr}{$start}++;
      $RCBKs{$chr2}{$end}++;
    }
  }
}

#Select all Self-Chain regions that overlap the transcripts that contain the breakpoint
foreach my $chr(keys %Rchrs){

  my @rposes=keys %{$RCBKs{$chr}} if(defined $RCBKs{$chr});
  foreach my $pos(@rposes){
    foreach my $e(@{$ABKs{$chr}{$pos}}){  #All the transcripts that overlap the breakpoint

      my @fstr=split /\./,$opts{C}; pop @fstr;
      my $ftab_chr=join(".", @fstr,"chr$chr",'tab');
      my $tStart;
      open(SC,"<$ftab_chr") || print STDERR "can't open $ftab_chr, skip annotating its self-chain\n" && next;
      do{
	$_=<SC>;
	chomp;
	next if(/^\#/);
	my $c;
	($c->{bin},$c->{score},$c->{tName},$c->{tSize},$c->{tStart},$c->{tEnd},$c->{qName},$c->{qSize},$c->{qStrand},$c->{qStart},$c->{qEnd},$c->{id},$c->{normScore})=split;
	$c->{tName}=~s/chr//;
	$c->{qName}=~s/chr//;

	my $size1=$c->{tEnd}-$c->{tStart};
	my $size2=$c->{qEnd}-$c->{qStart};
	next unless($c->{tName} eq $chr
		    || $c->{normScore}<$opts{h}
		    || $size1>$opts{z}
		    || $size2>$opts{z}
		   );
	if($c->{tEnd}>=$e->{txStart} && $c->{tStart}<=$e->{txEnd}){  #overlap by at least 1 bp
	  push @{$SCTBKs{$chr}{$pos}},$c;
	}
	$tStart=$c->{tStart};
      } until($tStart>$e->{txEnd} || eof(SC));
      close(SC);
    }
  }
}

if(@confirmed){
  printf "#%s %s %s\n",$version,$options,join(" ",@ARGV);
  printf "#Chr1\tPos1\tChr2\tPos2\tSource\tGeneFusion\tSelfChainScore\tGenomicAllele\n";
  foreach my $sp(@confirmed){
    my ($chr,$start,$chr2,$end,$source,@extra)=split /\s+/,$sp;
    $chr=~s/chr//;
    $chr2=~s/chr//;
    my $selfchainAnnot;
    if(defined $opts{C}){
      $selfchainAnnot=&GetSelfChainAnnotation($chr,$start,$chr2,$end);
    }
    if(defined $selfchainAnnot && $selfchainAnnot<$opts{h}){
      printf "%s\t%d\t%s\t%d\t%s\t%s\t%s\t%s\n",$chr,$start,$chr2,$end,$source,$fusion{$sp},$selfchainAnnot,$gBKPath{$sp};
      #printf "%s\t%s\t%s\n",$sp,$fusion{$sp},$gBKPath{$sp};
    }
  }
}

sub SimplePath{
  #convert a path into a string representation
  my ($sbp,$p)=@_;
  my @bps=($sbp,split /\|/,$p);
  my @np=($bps[0]);
  my $pathstr=$bps[0];
  if($#bps>=1){
    for(my $i=1;$i<$#bps;$i++){
      my ($pbp,$pori)=($bps[$i-1]=~/(\S+\:\d+)([\+\-])/);
      my ($bp,$ori)=($bps[$i]=~/(\S+\:\d+)([\+\-])/);
      my ($nbp,$nori)=($bps[$i+1]=~/(\S+\:\d+)([\+\-])/);
      if(defined $edges{$pbp}{$bp}{$pori.$ori} && $edges{$pbp}{$bp}{$pori.$ori}==0 &&
	 defined $edges{$bp}{$nbp}{$ori.$nori} && $edges{$bp}{$nbp}{$ori.$nori}==0
	){  #redundant normal adjacency
	#skip
      }
      else{
	if($np[$#np] ne $bps[$i]){  #non-redundant
	  if($#np==0 && defined $edges{$pbp}{$bp}{$pori.$ori} && $edges{$pbp}{$bp}{$pori.$ori}==0){  #head normal adjacency ignored
	    $pathstr=$bps[$i];
	    @np=($bps[$i]);
	  }
	  else{
	    $pathstr.=join("",$edges{$pbp}{$bp}{$pori.$ori}, '|', $bps[$i]);  #with read counts
	    push @np,$bps[$i];
	  }
	}
      }
    }
    if($np[$#np] ne $bps[$#bps]){  #last one
      my ($pbp,$pori)=($np[$#np]=~/(\S+\:\d+)([\+\-])/);
      my ($bp,$ori)=($bps[$#bps]=~/(\S+\:\d+)([\+\-])/);
      $pathstr.=join("", $edges{$pbp}{$bp}{$pori.$ori}, '|', $bps[$#bps]) if(defined $edges{$pbp}{$bp}{$pori.$ori} && $edges{$pbp}{$bp}{$pori.$ori}>0);
    }
  }

  return $pathstr;
}

sub RevOri{
  my ($ori)=@_;
  if($ori eq '+'){
    return '-';
  }
  else{
    return '+';
  }
}

sub getPath{
  #recursion
  my ($edges,$visited,$strand1,$bp1,$end,$strand2)=@_;
  my @path;

  if(defined $bp1 && defined $end && defined $$edges{$bp1}){
    my @bp2s=keys %{$$edges{$bp1}};
    if(@bp2s){
      foreach my $bp2(@bp2s){
	next unless (defined $bp2);
	foreach my $ori(keys %{$edges{$bp1}{$bp2}}){
	  my ($ori1,$ori2)=split //,$ori;
	  next unless($ori1 eq $strand1);
	  next if($$visited{$bp1}{$bp2}{$ori});  #avoid repeat
	  $$visited{$bp1}{$bp2}{$ori}=1;

	  if($bp2 eq $end && $ori2 eq $strand2){  #reached target
	    push @path, $end . $strand2;
	  }
	  else{
	    my @spath=&getPath($edges,$visited,$ori2,$bp2,$end,$strand2);
	    foreach my $p(@spath){
	      push @path, $bp2 . $ori2 . '|' . $p;
	    }
	  }
	}
      }
    }
  }
  return @path;
}

sub GetInnerDNAbp{
  my ($RNAbp,$RNApos)=@_;
  my $DNAbp;
  if(defined $RNAbp && $GeneBKs{$RNAbp->{gene}}){
    my ($chr,$pos);
    my @loci=@{$GeneBKs{$RNAbp->{gene}}};
    if($RNAbp->{strand} eq '+'){
      for(my $i=0;$i<=$#loci;$i++){  # innermost bp
	($chr,$pos)=split /\:/,$loci[$i];
	last if($pos>=$RNApos);
      }
    }
    else{
      for(my $i=$#loci;$i>=0;$i--){  # innermost bp
	($chr,$pos)=split /\:/,$loci[$i];
	last if($pos<=$RNApos);
      }
    }
    $DNAbp=join(":",$chr,$pos);
  }
  return $DNAbp;
}

sub GetBKGeneAnnotation{
  my ($chr,$start)=@_;

  #Annotate Breakpoint
  my @es;
  if(defined $ABKs{$chr}{$start}){
    @es=@{$ABKs{$chr}{$start}};
  }

  #select the transcript (from multiple ones)
  my $e;
  for(my $i=0;$i<=$#es;$i++){
    if(!defined $e || $e->{txSize} < $es[$i]->{txSize}){  #longest isoform
      $e=$es[$i];
    }
  }
  my $annot=&AnnotExon($start,$e) if(defined $e);
  return $annot;
}

sub AnnotExon{
  my ($pos,$e)=@_;
  my @exonStarts=split /\,/,$e->{exonStarts};
  my @exonEnds=split /\,/,$e->{exonEnds};

  my $report;
  ($report->{gene},$report->{strand})=($e->{name2},$e->{strand});
  for(my $i=0;$i<=$#exonStarts;$i++){
    if($exonStarts[$i]<=$pos && $pos<=$exonEnds[$i]){
      my $id=$i+1;
      my $rpos=$pos-$exonStarts[$i]+1;
      if($e->{strand} eq '-'){
	$id=$#exonStarts-$i+1;
	$rpos=$exonEnds[$i]-$pos+1;
      }
      ($report->{unit},$report->{id},$report->{pos},$report->{size})=('Exon',$id,$rpos,$exonEnds[$i]-$exonStarts[$i]+1);
    }
    elsif($i>0 && $exonEnds[$i-1]<$pos && $pos<$exonStarts[$i]){
      my $id=$i;
      my $rpos=$pos-$exonEnds[$i-1]+1;
      if($e->{strand} eq '-'){
	$id=$#exonStarts-$i+1;
	$rpos=$exonStarts[$i]-$pos+1;
      }
      ($report->{unit},$report->{id},$report->{pos},$report->{size})=('Intron',$id,$rpos,$exonStarts[$i]-$exonEnds[$i-1]+1);
    }
  }
  return $report;
}

sub ReadUCSCGeneAnnotation{
  my ($chr)=@_;
  my %Annot;
  my %Gene_starts;
  my %Gene_ends;
  open(AN,"<$opts{g}") || die "Unable to open $opts{g}\n";
  while(<AN>){
    chomp;
    next if(/^\#/);
    my $e;
    ($e->{bin},$e->{name},$e->{chrom},$e->{strand},$e->{txStart},$e->{txEnd},$e->{cdsStart},$e->{cdsEnd},$e->{exonCount},$e->{exonStarts},$e->{exonEnds},$e->{id},$e->{name2},$e->{cdsStartStat},$e->{cdsEndStat},$e->{exonFrames})=split;
    $e->{chrom}=~s/chr//;
    next unless($e->{chrom} eq $chr);
    $e->{txSize}=abs($e->{txStart}-$e->{txEnd}+1);
    push @{$Annot{$e->{chrom}}{$e->{txEnd}}{$e->{txStart}}},$e;
    my $start=($e->{strand} eq '+')?$e->{txStart}:$e->{txEnd};
    $Gene_starts{$e->{strand}}{$start}=$e->{name2};
    my $end=($e->{strand} eq '+')?$e->{txEnd}:$e->{txStart};
    $Gene_ends{$e->{strand}}{$end}=$e->{name2};
  }
  close(AN);

  return \%Annot;
}

sub GetSelfChainAnnotation{
  my ($chr,$start,$chr2,$end)=@_;
  my $chained=0;
  if(defined $SCTBKs{$chr}{$start} && defined $SCTBKs{$chr2}{$end}){
    foreach my $sc1(@{$SCTBKs{$chr}{$start}}){
      my ($qpos1,$qpos2)=($sc1->{qStart},$sc1->{qEnd});
      if($sc1->{qStrand} eq '-'){
	($qpos1,$qpos2)=($sc1->{qSize}-$sc1->{qEnd}+1,$sc1->{qSize}-$sc1->{qStart});
      }
      if($chr2 eq $sc1->{qName} && $end>=$qpos1-1000 && $end<=$qpos2+1000){
	$chained=($chained<$sc1->{normScore})?$sc1->{normScore}:$chained;
      }
    }

    foreach my $sc2(@{$SCTBKs{$chr2}{$end}}){
      my ($qpos1,$qpos2)=($sc2->{qStart},$sc2->{qEnd});
      if($sc2->{qStrand} eq '-'){
	($qpos1,$qpos2)=($sc2->{qSize}-$sc2->{qEnd}+1,$sc2->{qSize}-$sc2->{qStart});
      }
      if($chr eq $sc2->{qName} && $start>=$qpos1-1000 && $start<=$qpos2+1000){
	$chained=($chained<$sc2->{normScore})?$sc2->{normScore}:$chained;
      }
    }
  }
  return $chained;
}
